/**
  *  \file planetscentral/host/schedule.cc
  *  \brief Host: Scheduler operations
  */

#include <time.h>
#include "planetscentral/host/schedule.h"
#include "planetscentral/dbclient/dbhash.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "cpluslib/format.h"


/** Pick daytime for a new game. In case the schedule for a host does
    not specify one, pick one which is at a (supposedly) idle hour.
    \param root "game:" root tree */
int32_t
pickDayTime(DbSubtree& root)
{
    DbHash hours(root.hashKey("hours"));
    int32_t bestHour = 6;
    int32_t bestLoad = 0;
    for (int32_t i = 0; i < 24; ++i) {
        int32_t hourToTest = (6 + i) % 24;
        int32_t loadAtThisTime = hours.getIntField(format("%d", hourToTest));
        if (i == 0 || loadAtThisTime < bestLoad) {
            bestHour = hourToTest;
            bestLoad = loadAtThisTime;
        }
    }
    ++hours.intField(format("%d", bestHour));
    return 60*bestHour;
}

/** Time scale. This contains the number of seconds in a minute.
    By scaling it down, the system can be accelerated for testing. */
static int time_scale = 60;

static const int MINUTES_PER_DAY = 60*24;


Schedule::Schedule()
    : dirty(true),
      type(Stopped),
      weekdays(),
      interval(3),
      daytime(6*60),
      hostEarly(true),
      hostDelay(30),
      hostLimit(360),
      condition(NoCondition),
      condTurnOrTime(0)
{ }

void
Schedule::setType(Type t)
{
    this->type = t;
    this->dirty = true;
}

void
Schedule::setWeekDays(SmallSet<int8_t> weekdays)
{
    this->weekdays = weekdays;
    this->dirty = true;
}

void
Schedule::setInterval(int interval)
{
    this->interval = interval;
    this->dirty = true;
}

void
Schedule::setDaytime(int daytime)
{
    this->daytime = daytime;
    this->dirty = true;
}

void
Schedule::setHostEarly(bool value)
{
    this->hostEarly = value;
    this->dirty = true;
}

void
Schedule::setHostDelay(int32_t minutes)
{
    this->hostDelay = minutes;
    this->dirty = true;
}

void
Schedule::setHostLimit(int32_t minutes)
{
    this->hostLimit = minutes;
    this->dirty = true;
}

void
Schedule::setCondition(Condition cond, int32_t arg)
{
    this->condition      = cond;
    this->condTurnOrTime = arg;
    this->dirty          = true;
}

Schedule::Type
Schedule::getType() const
{
    return type;
}

SmallSet<int8_t>
Schedule::getWeekDays() const
{
    return weekdays;
}

int
Schedule::getInterval() const
{
    return interval;
}

int
Schedule::getDaytime() const
{
    return daytime;
}

bool
Schedule::getHostEarly() const
{
    return hostEarly;
}

int32_t
Schedule::getHostDelay() const
{
    return hostDelay;
}

int32_t
Schedule::getHostLimit() const
{
    return hostLimit;
}

Schedule::Condition
Schedule::getCondition() const
{
    return condition;
}

int32_t
Schedule::getConditionArg() const
{
    return condTurnOrTime;
}

bool
Schedule::isExpired(int32_t turn, int32_t time) const
{
    switch (condition) {
     case TurnCondition:
        // "Until turn 20" includes turn 20
        return turn > condTurnOrTime;
     case TimeCondition:
        // "Until time X" means it is discarded at time X, because the
        // next time we'll look at it it will be later.
        return time >= condTurnOrTime;
     default:
        return false;
    }
}

void
Schedule::loadFrom(DbHash h)
{
    type = Type(h.getIntField("type"));
    if (type == Weekly) {
        weekdays = SmallSet<int8_t>::fromInteger(h.getIntField("weekdays"));
    }
    if (type == Daily) {
        interval = h.getIntField("interval");
    }
    daytime   = h.getIntField("daytime");
    hostEarly = h.getIntField("hostEarly");
    hostDelay = h.getIntField("hostDelay");
    hostLimit = h.getIntField("hostLimit");
    condition = Condition(h.getIntField("condition"));
    if (condition == TurnCondition) {
        condTurnOrTime = h.getIntField("condTurn");
    }
    if (condition == TimeCondition) {
        condTurnOrTime = h.getIntField("condTime");
    }
    dirty = false;
}

void
Schedule::saveTo(DbHash h)
{
    h.setField("type", type);
    if (type == Weekly) {
        h.setField("weekdays", weekdays.toInteger());
    }
    if (type == Daily) {
        h.setField("interval", interval);
    }
    h.setField("daytime", daytime);
    h.setField("hostEarly", hostEarly);
    h.setField("hostDelay", hostDelay);
    h.setField("hostLimit", hostLimit);
    h.setField("condition", condition);
    if (condition == TurnCondition) {
        h.setField("condTurn", condTurnOrTime);
    }
    if (condition == TimeCondition) {
        h.setField("condTime", condTurnOrTime);
    }
    dirty = false;
}

/** Get next host. Returns a time > now for the next host. */
int32_t
Schedule::getNextHost(int32_t now) const
{
    switch (type) {
     case Stopped:
     case Quick:
     case Manual:
        return 0;

     case Daily:
        // Host happens precisely interval days after the previous host.
        // If host was delayed too much, move it farther a day.
        {
            int32_t prev = getPreviousHost(now);
            int32_t next = prev + interval*MINUTES_PER_DAY;
            if (now - prev > hostLimit) {
                // Host was delayed too much, so skip a day.
                next += MINUTES_PER_DAY;
            }
            return next;
        }

     case Weekly:
        {
            // Find next day.
            int32_t prev = getPreviousHost(now);
            int32_t day = ((now - daytime) / MINUTES_PER_DAY);
            for (int i = 0; i < 7; ++i) {
                ++day;
                int32_t weekday = getWeekdayFromTime(MINUTES_PER_DAY * day);
                if (weekdays.contains(weekday)) {
                    int32_t thisDate = daytime + MINUTES_PER_DAY * day;
                    if (now - prev > hostLimit) {
                        // Host was delayed too much, so skip this one
                        prev = thisDate;
                    } else {
                        // Accept
                        return thisDate;
                    }
                }
            }
        }
    }
    return 0;
}

/** Get previous host. Returns a time <= now. */
int32_t
Schedule::getPreviousHost(int32_t now) const
{
    switch (type) {
     case Stopped:
     case Quick:
     case Manual:
        return 0;

     case Daily:
        // Hosts can happen each day. So round up down to preceding occasion of /daytime/.
        return daytime + MINUTES_PER_DAY * ((now - daytime) / MINUTES_PER_DAY);

     case Weekly:
        // Find a day before now where this host could have happened.
        {
            int32_t day = ((now - daytime) / MINUTES_PER_DAY);
            for (int i = 0; i < 7; ++i) {
                int weekday = getWeekdayFromTime(MINUTES_PER_DAY * day);
                if (weekdays.contains(weekday)) {
                    break;
                }
                --day;
            }
            return daytime + MINUTES_PER_DAY * day;
        }
    }
    return 0;
}

/** Get first possile host before a date. Used after discontinuities in the schedule.
    If a weekly/daily schedule follows a Quick/Manual schedule that hosted
    rarely, getNextHost(lastHostTime) will usually return currentTime.
    However, players expect the host to run at the set time. Therefore,
    we fake a lastHostTime that causes the next host to be generated at a
    sufficiently sensible day. */
int32_t
Schedule::getPreviousVirtualHost(int32_t now) const
{
    // If now is at or slightly after a sensible host run, this would otherwise
    // return a date today for the virtual host run, setting the next one far
    // into the future. Thus, go a little back in time, using hostDelay as the
    // maximum allowed host delay. This will set the previous virtual host to
    // the previous date, placing the one following it again close to now, which
    // is what we want.
    now -= hostLimit;
    now -= 1;

    switch (type) {
     case Stopped:
     case Quick:
     case Manual:
        return 0;

     case Daily:
        // Hosts can happen each day. getPreviousHost() returns yesterday.
        // To get a host run for today, we'd have to add one day; the
        // virtual host would have to happen n-1 days before.
        return getPreviousHost(now) - (interval-1) * MINUTES_PER_DAY;

     case Weekly:
        // Find a day before now where this host could have happened. This happens
        // to be the same as getPreviousHost().
        return getPreviousHost(now);
    }
    return 0;
}

/** Get current time. We store MINUTES since epoch, giving us a little longer
    than 2038 if time() works right.

    This is therefore a minutes counter since Thu Jan 1 1970, 0:00.
    To obtain minutes: %60.
    To obtain hours: /60%24. */
int32_t
getCurrentTime()
{
    return time(0) / time_scale;
}

/** Get weekday. This assumes day 0 is Sunday. */
int
getWeekdayFromTime(int32_t time)
{
    /* Counting starts at a Thursday, which we want to be day 4. */
    return ((time / (24*60)) + 4) % 7;
}

time_t
getSystemTimeFromTime(int32_t t)
{
    return t*time_scale;
}

/** Convert time. In the protocol, times are always given in minutes-since-epoch.
    If the system internally runs on a higher time scale for testing, we have to
    convert them to not confuse users too much (users scale by 60 to obtain Unix
    time). */
int32_t
getUserTimeFromTime(int32_t t)
{
    return t*time_scale/60;
}

void
setTimeScale(int scale)
{
    time_scale = scale;
}
