/**
  *  \file planetscentral/rater/rater.cc
  *  \brief Difficulty Rating: Main Entry Point
  */

#include <cstring>
#include <cstdlib>
#include "planetscentral/rater/difficulty.h"
#include "arch/platform.h"
#include "util/console.h"
#include "util/modinit.h"
#include "cpluslib/format.h"
#include "util/except.h"
#include "game/formula.h"

static const char* progname;

static void
die(string_t s)
{
    console.write(LOG_ERROR, format("%s: %s", progname, s));
    std::exit(1);
}

static void
help()
{
    console << "Usage: " << progname << " [-options...] [gamedir]\n\n"
        "Options:\n"
        " -DSECTION.OPTION=VALUE    process option\n"
        " -file=FILE                process config file fragment for options\n"
        " -total                    only show total value as integer\n"
        " -value=WHAT               only show one sub-rating as integer,\n"
        "                           WHAT is shiplist, minerals, natives, production\n"
        "\n"
        "Default is a report with all sub-ratings and totals.\n"
        "A game directory is processed as a whole (config files and shiplist).\n";
    std::exit(0);
}

static void
showRating(DifficultyRater& rater, const char* ttl, DifficultyRater::Rating whichone)
{
    double result = rater.getRating(whichone);
    console.write(LOG_INFO, format("%-30s %6.2f%s", ttl, 100*result, rater.isRatingKnown(whichone) ? "" : " (default)"));
}

static const ModuleInitializer* rater_modules[] = {
    mod_base_console,
    0
};

int main(int /*argc*/, char** argv)
{
    /* Initialize */
    performInitialisation(OnStartup, rater_modules);
    argv = Platform::convertArgumentsToUtf8(argv);
    progname = argv[0];

    /* Parse arguments and operate difficulty rater */
    try {
        DifficultyRater rater;
        enum { Report, OneOnly, TotalOnly } output = Report;
        DifficultyRater::Rating output_one = DifficultyRater::ShiplistRating;
        while (const char* p = *++argv) {
            if (p[0] == '-') {
                if (p[1] == '-') {
                    ++p;
                }
                if (std::strcmp(p, "-help") == 0 || std::strcmp(p, "-h") == 0) {
                    // help
                    help();
                } else if (std::strncmp(p, "-D", 2) == 0) {
                    // direct assignment
                    p += 2;
                    const char* q = std::strchr(p, '=');
                    if (q == 0) {
                        die("\"-D\" option must have format \"-DNAME=VALUE\"");
                    } else {
                        rater.addConfig(string_t(p, q-p), q+1);
                    }
                } else if (std::strcmp(p, "-total") == 0) {
                    // report only total value
                    output = TotalOnly;
                } else if (std::strncmp(p, "-value=", 7) == 0) {
                    // report only one sub-rating
                    output = OneOnly;
                    const char* val = p+7;
                    std::size_t vlen = std::strlen(val);
                    if (vlen == 0) {
                        die("\"-value=\" needs an argument");
                    } else if (std::strncmp(val, "shiplist", vlen) == 0) {
                        output_one = DifficultyRater::ShiplistRating;
                    } else if (std::strncmp(val, "minerals", vlen) == 0) {
                        output_one = DifficultyRater::MineralRating;
                    } else if (std::strncmp(val, "natives", vlen) == 0) {
                        output_one = DifficultyRater::NativeRating;
                    } else if (std::strncmp(val, "production", vlen) == 0) {
                        output_one = DifficultyRater::ProductionRating;
                    } else {
                        die("invalid argument to \"-value=\"");
                    }
                } else if (std::strncmp(p, "-file=", 6) == 0) {
                    // read a config file
                    try {
                        Ptr<Stream> s = Stream::openFile(p+6, Stream::C_READ);
                        rater.addConfigFile(*s);
                    }
                    catch (FileProblemException& e) {
                        die(e.getFileName() + ": " + e.what());
                    }
                } else {
                    die("invalid option");
                }
            } else {
                // read a directory
                Ptr<Directory> dir = Directory::openDir(p);
                rater.addConfigDirectory(*dir);
                rater.addShipList(*dir);
            }
        }
    
        /* Result */
        switch (output) {
         case Report:
            showRating(rater, "Ship list rating",  DifficultyRater::ShiplistRating);
            showRating(rater, "Mineral rating",    DifficultyRater::MineralRating);
            showRating(rater, "Native rating",     DifficultyRater::NativeRating);
            showRating(rater, "Production rating", DifficultyRater::ProductionRating);
            console.write(LOG_INFO, format("%-30s %6.2f", ">> Total rating", 100*rater.getTotalRating()));
            break;
         case OneOnly:
            console.write(LOG_INFO, format("%d", roundToInt(100*rater.getRating(output_one))));
            break;
         case TotalOnly:
            console.write(LOG_INFO, format("%d", roundToInt(100*rater.getTotalRating())));
            break;
        }
    }
    catch (std::exception& e) {
        die(e.what());
    }
    return 0;
}
