/**
  *  \file planetscentral/talk/smiley.cc
  *  \brief Talk: Smileys
  */

#include <cstddef>
#include <cstring>
#include <cctype>
#include "planetscentral/talk/smiley.h"
#include "cpluslib/misc.h"
#include "cpluslib/types.h"

static const Smiley SMILEYS[] = {
    { "cool",       "B-)",  0,    "res/smileys/cool.png",       16, 16 },
    { "cry",        0,      0,    "res/smileys/cry.png",        16, 16 },
    { "eek",        0,      0,    "res/smileys/eek.png",        16, 16 },
    { "embarassed", 0,      0,    "res/smileys/embarassed.png", 16, 16 },
    { "evil",       ">:-)", 0,    "res/smileys/evil.png",       16, 16 },
    { "facepalm",   0,      0,    "res/smileys/facepalm.png",   16, 16 },
    { "frown",      0,      0,    "res/smileys/sad.png",        16, 16 }, /* same as sad */
    { "innocent",   0,      0,    "res/smileys/innocent.png",   16, 16 },
    { "lol",        ":-D",  ":D", "res/smileys/lol.png",        16, 16 },
    { "neutral",    ":-|",  0,    "res/smileys/neutral.png",    16, 16 },
    { "pacman",     0,      0,    "res/smileys/pacman.png",     16, 16 },
    { "rolleyes",   "8-|",  0,    "res/smileys/rolleyes.png",   16, 16 },
    { "sad",        ":-(",  ":(", "res/smileys/sad.png",        16, 16 },
    { "smile",      ":-)",  ":)", "res/smileys/smile.png",      16, 16 },
    { "tongue",     ":-P",  ":P", "res/smileys/tongue.png",     16, 16 },
    { "transversalis", 0,   0,    "res/smileys/transversalis.png", 32, 16 },
    { "wink",       ";-)",  ";)", "res/smileys/wink.png",       16, 16 },
};


/** Look up smiley by name.
    \param name Name given by user
    \return Pointer to smiley definition. NULL if not a valid smiley. */
const Smiley*
lookupSmiley(const string_t& name)
{
    for (std::size_t i = 0; i < countof(SMILEYS); ++i) {
        if (name == SMILEYS[i].name) {
            return &SMILEYS[i];
        }
    }
    return 0;
}

/*************************** findInlineElement ***************************/

static string_t
findFirstCharacters()
{
    string_t firsts = ":@";         /* colon is anchor point for URLs and :smileys:, @ is for email */
    for (std::size_t i = 0; i < countof(SMILEYS); ++i) {
        const Smiley& sm = SMILEYS[i];
        if (sm.symbol != 0 && firsts.find(sm.symbol[0]) == string_t::npos) {
            firsts += sm.symbol[0];
        }
        if (sm.symbol2 != 0 && firsts.find(sm.symbol2[0]) == string_t::npos) {
            firsts += sm.symbol2[0];
        }
    }
    return firsts;
}

static bool
findInlineURL(const string_t& text,
              string_t::size_type start_at,
              string_t::size_type anchor,
              InlineInfo& info)
{
    /* Protocols that are auto-detected */
    static const char*const PROTOCOLS[] = {
        "ftp",
        "http",
        "https",
        "mailto",
        "news",
        "nntp",
    };

    /* Check for protocol string */
    string_t::size_type n = anchor - start_at;
    string_t::size_type found_plen = 0;
    for (std::size_t i = 0; i < countof(PROTOCOLS); ++i) {
        string_t::size_type plen = std::strlen(PROTOCOLS[i]);
        if (plen <= n && text.compare(anchor - plen, plen, PROTOCOLS[i], plen) == 0) {
            found_plen = plen;
            break;
        }
    }
    if (found_plen == 0) {
        return false;
    }
    string_t::size_type url_start = anchor - found_plen;
    string_t::size_type url_end   = anchor + 1;

    /* Check whatever-before-the-string: must not be a letter */
    if (url_start > start_at && std::isalnum(uint8_t(text[url_start-1]))) {
        return false;
    }

    /* If it's an angle bracket, try to take that as delimiter */
    if (url_start > start_at && text[url_start-1] == '<') {
        while (url_end < text.size() && text[url_end] != '>' && text[url_end] != '\n') {
            ++url_end;
        }
        if (url_end < text.size() && text[url_end] == '>') {
            info.what   = InlineInfo::Link;
            info.start  = url_start;
            info.length = url_end - url_start;
            info.text.assign(text, url_start, url_end - url_start);
            return true;
        }
    }

    /* First character after the protocol must be slash or alphanumeric.
       This might fail some elaborate naming schemes, but will work for
       99.9% of all URLs out there, and users can still use [url] for
       those where it doesn't. */
    if (url_end >= text.size() || (text[url_end] != '/' && !std::isalnum(uint8_t(text[url_end])))) {
        return false;
    }

    /* Simple heuristic: scan until whitespace found. Honor parens. */
    int paren = 0;
    bool back_out = true;
    while (url_end < text.size()) {
        char c = text[url_end];
        if (c == '\n' || c == ' ' || c == '\t' || c == '\r') {
            /* URL ends at whitespace */
            back_out = true;
            break;
        } else if (c == '"' || c == '>') {
            /* URL ends at unlikely characters */
            back_out = false;
            break;
        } else if (c == '(') {
            /* Start paren. For Wikipedia ("http://.../wiki/Foo_(Bar)") and MSDN ("...(VS=8.1).aspx") URLs */
            ++paren;
        } else if (c == ')') {
            /* End paren. */
            if (paren == 0) {
                back_out = false;
                break;
            } else {
                --paren;
            }
        } else {
            /* Proceed */
        }
        ++url_end;
    }

    /* Make sure we don't include final punctuation */
    if (back_out && url_end > 0) {
        char prev = text[url_end-1];
        if (prev == '.' || prev == ',' || prev == ';' || prev == ':') {
            --url_end;
        }
    }

    /* Found. */
    info.what   = InlineInfo::Link;
    info.start  = url_start;
    info.length = url_end - url_start;
    info.text.assign(text, url_start, url_end - url_start);
    return true;
}

static bool
matchNamedSmiley(const string_t& text,
                 string_t::size_type anchor,
                 const char* name)
{
    string_t::size_type n = std::strlen(name);
    return text.size() - anchor >= n+2
        && text[anchor] == ':'
        && text[anchor + n + 1] == ':'
        && text.compare(anchor+1, n, name, n) == 0;
}

static bool
matchSymbolSmiley(const string_t& text,
                 string_t::size_type anchor,
                 const char* name)
{
    string_t::size_type n = std::strlen(name);

    /* Match smiley */
    if (text.size() - anchor < n || text.compare(anchor, n, name, n) != 0) {
        return false;
    }

    /* Match boundaries: a smiley starting with a letter must not immediately follow a letter,
       a smiley ending with a letter must not immediately be followed by a letter. */
    if (std::isalnum(uint8_t(name[0])) && anchor > 0 && std::isalnum(uint8_t(text[anchor-1]))) {
        return false;
    }

    if (std::isalnum(uint8_t(name[n-1])) && text.size() - anchor > n && std::isalnum(uint8_t(text[anchor + n]))) {
        return false;
    }

    return true;
}

static bool
findInlineSmiley(const string_t& text,
                 string_t::size_type /*start_at*/,
                 string_t::size_type anchor,
                 InlineInfo& info)
{
    /* Preinitialize info, assuming success */
    info.what  = InlineInfo::Smiley;
    info.start = anchor;

    /* Check all smileys */
    for (std::size_t i = 0; i < countof(SMILEYS); ++i) {
        const Smiley& sm = SMILEYS[i];
        if (matchNamedSmiley(text, anchor, sm.name)) {
            info.length = std::strlen(sm.name) + 2;
            info.text = sm.name;
            return true;
        }
        if (sm.symbol != 0 && matchSymbolSmiley(text, anchor, sm.symbol)) {
            info.length = std::strlen(sm.symbol);
            info.text = sm.name;
            return true;
        }
        if (sm.symbol2 != 0 && matchSymbolSmiley(text, anchor, sm.symbol2)) {
            info.length = std::strlen(sm.symbol2);
            info.text = sm.name;
            return true;
        }
    }
    return false;
}


/** Find an inline element.
    \param text     [in] Text to scan
    \param start_at [in] Start scanning at this position
    \param what     [in] What to find
    \param info    [out] Result
    \retval true Found an element, info has been updated
    \retval false Not found, info has indeterminate content */
bool
findInlineElement(const string_t& text,
                  string_t::size_type start_at,
                  InlineItemSet what,
                  InlineInfo& info)
{
    /* Quick exit: if we're not expected to return anything, don't bother searching */
    if (what.empty()) {
        return false;
    }

    /* List of possible first characters; initialized from smiley table */
    static const string_t FIRSTS = findFirstCharacters();

    while (1) {
        /* Look for anchor */
        string_t::size_type anchor = text.find_first_of(FIRSTS, start_at);
        if (anchor == string_t::npos) {
            return false;
        }

        /* Look for URLs */
        if (what.contains(InlineInfo::Link) && text[anchor] == ':') {
            if (findInlineURL(text, start_at, anchor, info)) {
                return true;
            }
        }

        /* Look for smileys */
        if (what.contains(InlineInfo::Smiley)) {
            if (findInlineSmiley(text, start_at, anchor, info)) {
                return true;
            }
        }

        /* Advance */
        start_at = anchor+1;
    }
}

#ifdef TEST_SMILEY
/*
  g++-3.3 -O -DTEST_SMILEY -o test_smiley -I. -Icpluslib planetscentral/talk/smiley.cc -L. -lcplus
 */
#include <iostream>

std::ostream& operator<<(std::ostream& os, InlineInfo::What w)
{
    switch (w) {
     case InlineInfo::Smiley: return os << "Smiley";
     case InlineInfo::Link:   return os << "Link";
    }
    return os << "Unknown(" << int(w) << ")";
}

int main()
{
    string_t line;
    while (getline(std::cin, line)) {
        string_t::size_type start_at = 0;
        InlineInfo info;
        while (findInlineElement(line, start_at, InlineItemSet::fromInteger(-1U), info)) {
            std::cout << "Text:        '" << line.substr(start_at, info.start - start_at) << "'\n";
            std::cout << "Inline elem: " << info.what << ", length=" << info.length << ", text='" << info.text << "'\n";
            start_at = info.start + info.length;
        }
        std::cout << "Final text:  '" << line.substr(start_at) << "'\n";
    }
}
#endif
